from tools import *


class MyLayer:
    """ Template example for the class implementing a single layer """
    
    def forward(self, x: np.array, w: np.array) -> np.array:
        """
        :param x: [N x d] -- layer input of N data points with d dimensions
        :param w: [d] -- d weights
        :return: layer output y = x * w
        """
        self.x = x
        self.w = w
        return x * w[np.newaxis, :]
    
    def backward(self, dy: np.array) -> Tuple[np.array, np.array]:
        """
        :param dy: [N x d] -- gradient with respect to the layer output
        :param w: [d] -- d weights
        :return: tuple of gradients in the layer inputs (dx, dw)
        """
        dx = dy * self.w[np.newaxis, :]
        dw = (dy * self.x).sum(axis=0)
        return (dx, dw)


class MyLoss:
    """ Template example for the loss function """
    
    def forward(self, x: np.array, y: np.array) -> float:
        """
        :param x: [N] -- scores of N data points
        :param y: [N] -- target labels
        """
        # regression
        self.delta = x - y
        return self.delta ** 2
    
    def backward(self) -> np.array:
        """
        :return: gradient in the layer input dx
        """
        dx = 2 * self.delta
        return dx


class MyNet:
    """ Template example for the network """
    
    def __init__(self, input_size, hidden_size):
        # this is needed for printing
        self.name = f'Net-example-hidden-{hidden_size}'
        # define some arrays that will be parameters of the network
        self.params = dotdict()  # same as dict but with additional access through dot notation
        self.params.a = np.random.rand(input_size)
        self.params.b = np.random.rand(input_size)
        # define some layers
        self.layer1 = MyLayer()
        self.layer2 = MyLayer()
        self.loss = MyLoss()
    
    def score(self, x: np.array) -> np.array:
        """
        Return log of predictive probability p(y|x) of the network
        *
        :param x: np.array [N x d], N number of points, d dimensionality of the input features
        :return: y: np.array [N] predicted scores of class 1 for all points
        """
        h = self.layer1.forward(x, self.params.a)
        y = self.layer2.forward(h, self.params.b)
        return y.sum(axis=1)
    
    def classify(self, x: np.array) -> np.array:
        """
        Make class prediction for the given gata
        *
        :param x: np.array [N x d], N number of points, d dimensionality of the input features
        :return: y: np.array [N] class 0 or 1 per input point
        """
        return self.score(x) > 0
    
    def mean_loss(self, x, y):
        """
        Compute the total loss on the training data
        *
        :param train_data: tuple(x,y)
        x [N x d] np.array data points
        y [N] np.array their classes
        :return: total loss to optimize
        """
        s = self.score(x)
        return self.loss.forward(s, y).mean()
    
    def backward(self):
        """
        Compute gradients in all parameters
        *
        :return: dict mapping parameters to their gradients for grad descent
        """
        grads = dotdict()
        score_grad = self.loss.backward()
        #
        # this is a backprop from the sum in self.score(), for which I was too lazy to define a layer
        # and look how ugly it becomes here
        score_grad = np.tile(score_grad[:, np.newaxis], (1, self.params.a.size)) / score_grad.size
        #
        # contunue with modular propagation
        hidden_grad, b_grad = self.layer2.backward(score_grad)
        grads.b = b_grad
        input_grad, a_grad = self.layer1.backward(hidden_grad)
        grads.a = a_grad
        return grads
    
    def train(self, train_data, epochs=100, step_size=0.1):
        """
        Train the model using gradient descent
        *
        :param train_data: tuple (x,y) of trianing data
        :param epochs: number of epochs for gradient descent
        :param step_size: step size in gradient descent
        """
        x, y = train_data
        for epoch in range(epochs):
            # compute loss of the train data
            L = self.mean_loss(x, y)
            # compute gradients in all parameter
            grads = self.backward()
            # make a grad descent step
            for (k, p) in self.params.items():
                p -= step_size * grads[k]
                # print current loss, should be going down
                print(L)


""" load model """
model_dict = load_object("model.pkl")
for (k, v) in model_dict.items():
    print(f'{k}:{v}')

# create our network
net = MyNet(model_dict['dim'], 2)

# use provided class to generate data, visualize data and decision boundaries
model = G2Model(**model_dict)
train_data = model.generate_sample(100)

print("Training net:")
net.train(train_data, epochs=100)

# get some test data
test_data = model.generate_sample(1000)
err_rate = model.test_error(net, test_data)
print(f"Achieved error rate:{err_rate * 100:.3f}%")

#
print(f"Drawing -- see {net.name}.pdf")
model.plot_predictor(net, train_data)
