from scipy.stats import norm
from scipy.special import expit
from scipy.stats import special_ortho_group
import scipy.stats
from typing import Tuple

import math
import numpy as np

import pickle

import os
import sys

""" matplotlib drawing to a pdf setup """
import matplotlib
matplotlib.use('Agg')
import matplotlib.pyplot as plt


def save_pdf(file_name):
    plt.savefig(file_name, bbox_inches='tight', dpi=199, pad_inches=0)


figsize = (6.0, 6.0 * 3 / 4)


def save_object(filename, obj):
    with open(filename, 'wb') as output:
        pickle.dump(obj, output, pickle.DEFAULT_PROTOCOL)


def load_object(filename):
    res = pickle.load(open(filename, "rb"))
    return res


""" Class representing GT model and some tools to evaluate and visualize predictors """


class G2Model:
    def __init__(self, dim, clps, mues, covs):
        self.dim = dim
        self.set_clps(clps)
        self.set_mues(mues)
        self.set_covs(covs)
    
    def set_covs(self, covs):
        self.covs = np.array(covs, dtype=np.float64)
        if self.covs.shape != (2, self.dim, self.dim):
            raise Exception("Wrong shape of covs")
        
        ok = True
        for i in (0, 1):
            if np.allclose(self.covs[i], self.covs[i].T):
                try:
                    np.linalg.cholesky(self.covs[i])
                except np.linalg.LinAlgError:
                    print("matrix not positive definite")
                    ok = False
            else:
                print("matrix not symmetric")
                ok = False
        if not ok:
            raise Exception("covs not symmetric or not positive definite")
    
    def set_mues(self, mues):
        self.mues = np.array(mues, dtype=np.float64)
        if self.mues.shape != (2, self.dim):
            raise Exception("Wrong shape of mues")
    
    def set_clps(self, probs):
        self.clpr = np.array(probs, dtype=np.float64)
        if self.clpr.shape != (2,):
            raise Exception("Shape mismatch")
        if not np.all(self.clpr > 0):
            raise Exception("class probabilities not strictly positive")
        self.clpr /= np.sum(self.clpr)
    
    def generate_sample(self, sample_size):
        """
        function to draw labeled samples from the model
        :param sample_size: how many in total
        :return: (x,y)
        """
        y = np.zeros(1)
        if sample_size < self.dim * 2:
            raise Exception("Need at least 2*dim points for non-degenerate MLE estimates")
        # resample until get at least one point in each class
        while (y == 0).sum() < self.dim or (y == 1).sum() < self.dim:
            y = (np.random.uniform(size=sample_size) > self.clpr[0]) * 1
        mask0 = y == 0
        mask1 = np.logical_not(mask0)
        x0 = np.random.multivariate_normal(self.mues[0], self.covs[0], size=np.sum(mask0))
        x1 = np.random.multivariate_normal(self.mues[1], self.covs[1], size=np.sum(mask1))
        x = np.zeros((sample_size, self.dim), dtype=np.float64)
        x[mask0, :] = x0
        x[mask1, :] = x1
        
        return x, y
    
    def test_error(self, predictor, test_data):
        """
        evaluate test error of a predictor
        :param predictor: object with predictor.classify(x:np.array) -> np.array
        :param test_data: tuple (x,y) of the test points
        :return: error rate
        """
        x, y = test_data
        y1 = predictor.classify(x)
        err_rate = (y1 != y).sum() / x.shape[0]
        return err_rate
    
    def plot_predictor(self, predictor, train_data):
        """
        Visualizes the GT model, training points and the decisison boundary of a given predictor
        Assuming mean values are centered, so that the origing already in the linear span of (mu0, mu1)
        predictor: object with
            predictor.classify(x:np.array) -> np.array
            predictor.name -- str to appear in the figure
        :param train_data: tuple (x,y)
        """
        x, y = train_data
        y1 = predictor.classify(x)
        err = y1 != y
        #
        plt.figure(2, figsize=figsize)
        plt.rc('lines', linewidth=1)
        # project data
        v0 = self.mues[1] - self.mues[0]
        v0 /= np.linalg.norm(v0)
        # cov matrix full data
        C = self.covs[0] * self.clpr[0] + self.covs[1] * self.clpr[1]
        # project Covariance matrix to to v0
        dims = v0.size
        P = np.eye(dims) - np.outer(v0, v0)
        C = np.matmul(np.matmul(P, C), P)
        w, V = np.linalg.eig(C)
        i = np.argsort(w)  # sort ascending
        v1 = V[:, i[-1]]
        P = np.stack([v0, v1], axis=1)  # direction between means and largest eigenvector from projected C
        # marginal model in this projection, currentyl not used
        pg_dict = dict(dim=2)
        pg_dict['clps'] = self.clpr
        pg_dict['mues'] = np.matmul(self.mues, P)
        pcovs = [np.matmul(P.transpose(), np.matmul(self.covs[i], P)) for i in range(2)]
        pg_dict['covs'] = np.stack(pcovs, axis=0)
        pg = G2Model(**pg_dict)
        # pg.contour_plot()
        px = np.matmul(x, P)
        # plot points
        mask0 = y == 0
        mask1 = y == 1
        plt.plot(px[mask0, 0], px[mask0, 1], 'bo', ms=3)
        plt.plot(px[mask1, 0], px[mask1, 1], 'rd', ms=3)
        plt.plot(px[err, 0], px[err, 1], 'ko', ms=6, fillstyle='none', label='errors', markeredgewidth=0.5)
        # plot classifier boundary
        ngrid = [200, 100]
        xx = [np.linspace(px[:, i].min(), px[:, i].max(), ngrid[i]) for i in range(2)]
        Xi, Yi = np.meshgrid(xx[0], xx[1], indexing='ij')  # 200 x 100 matrices
        # need to map these points back to Rd to present to the classifier
        X = np.outer(Xi.flatten(), v0) + np.outer(Yi.flatten(), v1)
        # Plot the GT scores contour
        scores = [scipy.stats.multivariate_normal(self.mues[i], self.covs[i]).logpdf(X).reshape(ngrid) + math.log(self.clpr[i]) for i in range(2)]
        score = scores[1] - scores[0]
        m1 = np.linspace(0, score.max(), 4)
        m2 = np.linspace(score.min(), 0, 4)
        plt.contour(Xi, Yi, score, np.sort(np.concatenate((m1[1:], m2[0:-1]))))
        CS = plt.contour(Xi, Yi, score, [0], colors='r', linestyles='dashed')
        CS.collections[0].set_label('GT boundary')
        Y = predictor.classify(X).reshape(ngrid).astype(float)
        # maybe blur Y with half the cell size for smoothness
        from scipy.ndimage.filters import gaussian_filter
        Y = gaussian_filter(Y, sigma=1)
        CS = plt.contour(Xi, Yi, Y, [0.5], colors='k', linewidths=1)
        CS.collections[0].set_label('Predictor boundary')
        plt.xlabel("Projected dim 0")
        plt.ylabel("Projected dim 1")
        plt.text(0.3, 1.0, predictor.name, ha='center', va='top', transform=plt.gca().transAxes)
        plt.legend(loc=0)
        save_pdf(f'{predictor.name}.pdf')
        plt.clf()


class dotdict(dict):
    """dot.notation access to dictionary attributes"""
    """ For a more elaborate solution take a look at the EasyDict package https://pypi.org/project/easydict/ """
    __getattr__ = dict.get
    __setattr__ = dict.__setitem__
    __delattr__ = dict.__delitem__
    

def implement_me(f):
    return f
